! ----------------------------------------------------------------------
! MESSAGE PASSING INTERFACE TEST CASE SUITE
! 
! Copyright - 1996 Intel Corporation
! 
! Intel Corporation hereby grants a non-exclusive license under Intel's
! copyright to copy, modify and distribute this software for any purpose 
! and without fee, provided that the above copyright notice and the
! following paragraphs appear on all copies.
! 
! Intel Corporation makes no representation that the test cases
! comprising this suite are correct or are an accurate representation
! of any standard.
! 
! IN NO EVENT SHALL INTEL HAVE ANY LIABILITY FOR ANY DIRECT, INDIRECT
! OR SPECULATIVE DAMAGES, (INCLUDING WITHOUT LIMITING THE FOREGOING,
! CONSEQUENTIAL, INCIDENTAL AND SPECIAL DAMAGES) INCLUDING, BUT NOT
! LIMITED TO INFRINGEMENT, LOSS OF USE, BUSINESS INTERRUPTIONS, AND
! LOSS OF PROFITS, IRRESPECTIVE OF WHETHER INTEL HAS ADVANCE NOTICE OF
! THE POSSIBILITY OF ANY SUCH DAMAGES.
! 
! INTEL CORPORATION SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING,
! BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
! FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.  THE SOFTWARE PROVIDED
! HEREUNDER IS ON AN "AS IS" BASIS AND INTEL CORPORATION HAS NO
! OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS OR
! MODIFICATIONS.
! 
! ----------------------------------------------------------------------
! **********************************************************************
!                          Test for MPI_Allgatherv()
! 
! This code tests the functionality of the MPI_Allgatherv() function.
! This function extends the functionality of MPI_Allgather() (to gather
! a piece of data from each node to every other node) so that the pieces
! of data from different nodes may be different sizes for different
! nodes, and so that the displacement from the beginning of the receive
! buffer may leave spaces in between sections of data.
! 
! The latter feature is useful if one is sending data stored as a matrix
! with a leading dimension.  In this case, the beginning of the ith
! column is at location buffer[0 + i*leading_dim], and one might want to
! store matrices whose columns' lengths are less than the leading
! dimension.  In that case, if one wanted to do an Allgather() of
! columns of the matrix, one would use MPI_Allgatherv().
! 
! This test uses the standard MPITEST library functions to generate the
! message lengths, communicator sizes and types, and buffer types.  Once
! the environment is set up, the variable leading_dim is set to length
! plus some constant value, then the buffers are allocated accordingly
! (send_buffer has size "length", and recv_buffer has size
! "test_nump*leading_dim".)  The receive buffer is initialized to either
! -1, 999, or 'z' (depending on the buffer type in use), and the send
! buffer is initialized to something unique to the (the process' rank in
! the case of the integer types, the rank plus a small real part in the
! case of floats and doubles, and a character representation of the rank
! in the case of the character types.
! 
! Several subsidiary arrays are necessary for this test.  The arrays
! recv_counts[] and recv_displs[] list the message length and starting
! displacement in the receive array of the different process's messages.
! The arrays counts[] and displs[] and values[] are used for error
! checking.  counts[2*i] is equal to recv_counts[i] (i.e. the length of
! the message from the ith process), but counts[2*i+1] is equal to the
! length of the space in the receive buffer between the message from the
! ith process and that from the (i+1)th process.  In this test,
! counts[2*i+1]=(leading_dim-length) for all 0 < i < test_nump.
! Similarly, displs[2*i] is equal to recv_displs[i] (the displacement of
! the ith message from the beginning of the receive buffer), and
! displs[2*i+1] holds the displacement of the space between the ith and
! (i+1)th messages.  The array values[] holds the correct values which
! should be in the receive buffer upon succesful completion of this
! test.  values[2*i] holds the same value as the send buffer on the ith
! process, and values[2*i+1] holds the value which was in the receive
! buffer before the call to MPI_Allgatherv().  Note that values[] is an
! array of type dataTemplate, so each element must be filled with all
! possible types of value before it is used.
!
! Revision History:
!  1          gt  Ported from C.
! **********************************************************************

#include "f90options.h"

      INCLUDE 'mpitest_cfgf90.h'

      PROGRAM MAIN
        USE MPI

      INCLUDE 'mpitestf90.h'
      INCLUDE 'externalf90.h'


      INTEGER    COMM_INDEX
!                               the array index of the current comm
      INTEGER    COMM_TYPE
!                               the index of the current communicator type
      INTEGER    COMM_COUNT
!                               loop counter for communicator loop
      INTEGER    COMM_SIZE
!                               number of ranks in the current communicator
      LOGICAL    INTER_FLAG
!                               used to test if intercommunicator
      INTEGER    TYPE_COUNT
!                               number of data types to test
      INTEGER    TEST_TYPE
!                               index of current data type
      INTEGER    LENGTH_COUNT
!                               number of data lengths to test
      INTEGER    MAX_LENGTH
!                               max message length to test
      INTEGER    MAX_BYTE_LENGTH
!                               max byte length to test
      INTEGER    MAX_BUFF_LENGTH
!                               max length that will fit in buffers
      INTEGER    BYTE_LENGTH
!                               current byte length to test 
      INTEGER    LENGTH
!                               current length to test 
      INTEGER    FAIL
      INTEGER    LOOP_CNT
!                               counts total number of failures, loops
      INTEGER    ERROR
!                               number of data errors found in buffer
      INTEGER    IERR
      INTEGER    ERR
!                               return value from MPI calls
      INTEGER    COMM
!                               MPI communicator under test
      INTEGER    I, J, K, L, M
      INTEGER    LEADING_DIM

      MPITEST_BUF_TYPE SEND_BUFFER(MAX_BUFF_SIZE+2*MAX_RANKS+4)
      MPITEST_BUF_TYPE RECV_BUFFER(MAX_BUFF_SIZE+2*MAX_RANKS+4)
      INTEGER    ERRSIZE

      MPITEST_BUF_TYPE VALUES(MAX_RANKS+MAX_RANKS)
      MPITEST_BUF_TYPE VALUE
      INTEGER    RECV_COUNTS(MAX_RANKS)
      INTEGER    RECV_DISPLS(MAX_RANKS)
      INTEGER    COUNTS(MAX_RANKS+MAX_RANKS)
      INTEGER    DISPLS(MAX_RANKS+MAX_RANKS)

      CHARACTER*(IOSIZE)  INFOBUF
      CHARACTER*32   TESTNAME
       CHARACTER*(MPI_MAX_ERROR_STRING) ERRORSTRING

 99   FORMAT(A,INT_FMT)
 179  FORMAT(A,I6,A,I6,A,I6,A,I3,A,INT_FMT)
 189  FORMAT(INT_FMT,A,A,I6,A,I6,A,I6,A,I3,A,INT_FMT)


!
!     Initialize the MPI environment
!
      CALL MPI_INIT(IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        WRITE(INFOBUF,99) 'MPI_INIT() returned', IERR
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF

!
!     Initialize the MPITEST environment
!
      CALL MPITEST_INIT(IERR)
      TESTNAME = 'MPI_Allgatherv()'
      IF (MPITEST_ME .EQ. 0) THEN
        INFOBUF = 'Starting test '//TESTNAME
        CALL MPITEST_MESSAGE(MPITEST_INFO0, INFOBUF)
      END IF


!
!     Initialize the loop/error counters
!
      LOOP_CNT = 0
      FAIL = 0

!
!     Get number of communicators, data types and message lengths to test
!
      CALL MPITEST_NUM_COMM_SIZES(COMM_COUNT)
      CALL MPITEST_NUM_DATATYPES(TYPE_COUNT)
      CALL MPITEST_NUM_MESSAGE_LENGTHS(LENGTH_COUNT)

!
!     Get max message length for test
!
      CALL MPITEST_GET_MAX_MESSAGE_LENGTH(MAX_BYTE_LENGTH)

!
!     Loop through communicators
!
      DO 100 I=1,COMM_COUNT
        CALL MPITEST_GET_COMM_INDEX(I, COMM_INDEX)
        CALL MPITEST_GET_COMM_TYPE(I, COMM_TYPE)

        CALL MPITEST_GET_COMMUNICATOR(COMM_TYPE, COMM_INDEX, COMM, &
              COMM_SIZE)

!       Skip to end of loop if this node is not part of current
!       communicator

        IF (COMM .EQ. MPI_COMM_NULL) GO TO 201

!       Skip if intercommunicator

        CALL MPI_COMM_TEST_INTER(COMM, INTER_FLAG, IERR)
        IF (INTER_FLAG .EQV. .TRUE.) GO TO 201

!       Loop through data types.
!
        DO 200 J=1,TYPE_COUNT

          CALL MPITEST_GET_DATATYPE(J, TEST_TYPE)
!
!         Convert the number of bytes in the maximum length message
!         into the number of elements of the current type 
          CALL MPITEST_BYTE_TO_ELEMENT(TEST_TYPE, MAX_BYTE_LENGTH, &
                MAX_LENGTH)
          CALL MPITEST_BYTE_TO_ELEMENT(TEST_TYPE, &
                MAX_BUFF_SIZE * MPITEST_BUF_EXTENT, MAX_BUFF_LENGTH)

          IF (MAX_LENGTH .GE. ((MAX_BUFF_LENGTH /  &
                                 MPITEST_NUMP)-4))  &
              MAX_LENGTH = (MAX_BUFF_LENGTH/MPITEST_NUMP)-4
          IF (MAX_LENGTH .LE. 0) MAX_LENGTH = 0

!
!         Loop through data lengths
!
          DO 300 K=1,LENGTH_COUNT
            CALL MPITEST_GET_MESSAGE_LENGTH(K, BYTE_LENGTH)
            CALL MPITEST_BYTE_TO_ELEMENT(TEST_TYPE, BYTE_LENGTH, &
                LENGTH)
            IF (LENGTH .GT. MAX_LENGTH) LENGTH = MAX_LENGTH

!
!           Initialize send data
!
            LEADING_DIM = 1 + LENGTH
            CALL MPITEST_DATATEMPLATE_INIT(VALUE, &
                  MPITEST_CURRENT_RANK, TEST_TYPE)
            CALL MPITEST_INIT_BUFFER(TEST_TYPE, LENGTH+1, &
                  VALUE, SEND_BUFFER)
!
!           Initialize receive buffer and template
!
            CALL MPITEST_DATATEMPLATE_INIT(VALUE, -1, TEST_TYPE)
            CALL MPITEST_INIT_BUFFER(TEST_TYPE, COMM_SIZE*LEADING_DIM, &
                  VALUE, RECV_BUFFER)
           
            M = 0
            DO 400 L=1,COMM_SIZE
              COUNTS(2*L-1)=LENGTH+M
              COUNTS(2*L) = LEADING_DIM - LENGTH - M
              DISPLS(2*L-1)=(L-1)*LEADING_DIM
              DISPLS(2*L)=DISPLS(2*L-1) + LENGTH + M
              RECV_COUNTS(L) = LENGTH + M
              RECV_DISPLS(L) = (L - 1)*LEADING_DIM
              CALL MPITEST_DATATEMPLATE_INIT_POS(VALUES, L-1, &
                   2*L - 1, TEST_TYPE)
              CALL MPITEST_DATATEMPLATE_INIT_POS(VALUES, -1, &
                   2*L, TEST_TYPE)
              M = 1 - M
 400        CONTINUE

!           Print an informational message
            IF (MPITEST_CURRENT_RANK .EQ. 0) THEN
              WRITE (INFOBUF,179) 'commindex',COMM_INDEX,', commsize', &
                  COMM_SIZE, ', commtype', COMM_TYPE,  &
                  ', datatype', TEST_TYPE, ', length',  LENGTH
              CALL MPITEST_MESSAGE(MPITEST_INFO1, INFOBUF)
            END IF

            LOOP_CNT = LOOP_CNT + 1

            M = MOD(MPITEST_CURRENT_RANK, 2)
            CALL MPI_ALLGATHERV(SEND_BUFFER, LENGTH+M, &
                 MPITEST_MPI_DATATYPES(TEST_TYPE), &
                 RECV_BUFFER, RECV_COUNTS, RECV_DISPLS, &
                 MPITEST_MPI_DATATYPES(TEST_TYPE), COMM, IERR)
            IF (IERR .NE. MPI_SUCCESS) THEN
              WRITE(INFOBUF,99) 'MPI_ALLGATHERV() returned', IERR
              CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
              INFOBUF=' '
              CALL MPI_ERROR_STRING(IERR, ERRORSTRING, ERRSIZE, ERR)
              CALL MPITEST_MESSAGE(MPITEST_NONFATAL, ERRORSTRING)
              FAIL = FAIL + 1
            END IF
!
!           Check received data for errors . . .
!
            CALL MPITEST_BUFFER_ERRORS_V(TEST_TYPE, 2*COMM_SIZE, &
                  COUNTS, DISPLS, VALUES, RECV_BUFFER, ERROR)

            IF (ERROR .NE. 0) THEN
              IF (IERR .EQ. MPI_SUCCESS) FAIL = FAIL + 1
              WRITE (INFOBUF,189) ERROR, ' errors in buffer, ', &
                  'commindex',COMM_INDEX,', commsize', &
                  COMM_SIZE, ', commtype', COMM_TYPE,  &
                  ', datatype', TEST_TYPE, ', length', LENGTH
              CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
            END IF

 300      CONTINUE

 200    CONTINUE
 201    CONTINUE

        CALL MPITEST_FREE_COMMUNICATOR(COMM_TYPE, COMM, IERR)

 100  CONTINUE

!
!     Report overall results
!
      CALL MPITEST_REPORT(LOOP_CNT - FAIL, FAIL, 0, TESTNAME)

      CALL MPI_FINALIZE(IERR)

      END  
  
  
