! ----------------------------------------------------------------------
! MESSAGE PASSING INTERFACE TEST CASE SUITE
! 
! Copyright - 1996 Intel Corporation
! 
! Intel Corporation hereby grants a non-exclusive license under Intel's
! copyright to copy, modify and distribute this software for any purpose 
! and without fee, provided that the above copyright notice and the
! following paragraphs appear on all copies.
! 
! Intel Corporation makes no representation that the test cases
! comprising this suite are correct or are an accurate representation
! of any standard.
! 
! IN NO EVENT SHALL INTEL HAVE ANY LIABILITY FOR ANY DIRECT, INDIRECT
! OR SPECULATIVE DAMAGES, (INCLUDING WITHOUT LIMITING THE FOREGOING,
! CONSEQUENTIAL, INCIDENTAL AND SPECIAL DAMAGES) INCLUDING, BUT NOT
! LIMITED TO INFRINGEMENT, LOSS OF USE, BUSINESS INTERRUPTIONS, AND
! LOSS OF PROFITS, IRRESPECTIVE OF WHETHER INTEL HAS ADVANCE NOTICE OF
! THE POSSIBILITY OF ANY SUCH DAMAGES.
! 
! INTEL CORPORATION SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING,
! BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
! FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.  THE SOFTWARE PROVIDED
! HEREUNDER IS ON AN "AS IS" BASIS AND INTEL CORPORATION HAS NO
! OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS OR
! MODIFICATIONS.
! 
! ----------------------------------------------------------------------
! **********************************************************************
!                           Test for MPI_Scatterv()
!
! The MPI_Scatterv() function takes a vector of data on one process
! (which we refer to as the 'root' process) and sends a small piece of
! the vector to each of the other processes.  The root process specifies
! a vector of counts and a vector of displacements.  The ith element of
! these arrays gives the number of data elements to be sent to the
! ith process and the displacement from the beginning of the send buffer
! from which to take the ith process's message, respectively.
! 
! This test initializes the send buffer with the root's rank (for parts
! of the buffer which will be sent) and with '-1's in the spaces
! between the send buffers which will be used.  Then the
! Scatterv is performed.  All processes then test their received data to make
! sure it contains the root's rank.  For a given communicator, the test
! loops over the identity of the root, so that each process in the
! communicator acts once as root.
! 
! The MPITEST environment provides looping over message length, data
! type, and communicator size and type, as specified in the file
! mpitest_cfgf.h 
!
! Revision History:
!  1          gt  Ported from C.
! **********************************************************************

#include "f90options.h"

      INCLUDE 'mpitest_cfgf90.h'

      PROGRAM MAIN
        USE MPI

      INCLUDE 'mpitestf90.h'

      INCLUDE 'externalf90.h'

      INTEGER    COMM_INDEX
!                               the array index of the current comm
      INTEGER    COMM_TYPE
!                               the index of the current communicator type
      INTEGER    COMM_COUNT
!                               loop counter for communicator loop
      INTEGER    COMM_SIZE
!                               number of ranks in the current communicator
      LOGICAL    INTER_FLAG
!                               used to test if intercommunicator
      INTEGER    TYPE_COUNT
!                               number of data types to test
      INTEGER    TEST_TYPE
!                               index of current data type
      INTEGER    LENGTH_COUNT
!                               number of data lengths to test
      INTEGER    MAX_LENGTH
!                               max message length to test
      INTEGER    MAX_BYTE_LENGTH
!                               max byte length to test
      INTEGER    MAX_BUFF_LENGTH
!                               max length that will fit in buffers
      INTEGER    BYTE_LENGTH
!                               current byte length to test 
      INTEGER    LENGTH
!                               current length to test 
      INTEGER    LEADING_DIM
!                               current length to test 
      INTEGER    ROOT
!                               the current scatter root
      INTEGER    FAIL
      INTEGER    LOOP_CNT
!                               counts total number of failures, loops
      INTEGER    ERROR, ERROR2
!                               number of data errors found in buffer
      INTEGER    IERR
      INTEGER    ERR
!                               return value from MPI calls
      INTEGER    COMM
!                               MPI communicator under test
      INTEGER    I, J, K, L, M

      MPITEST_BUF_TYPE SEND_BUFFER(MAX_BUFF_SIZE+2*MAX_RANKS)
      MPITEST_BUF_TYPE RECV_BUFFER(MAX_BUFF_SIZE+2*MAX_RANKS)
      INTEGER    ERRSIZE

      MPITEST_BUF_TYPE VALUES(MAX_RANKS+MAX_RANKS)
      MPITEST_BUF_TYPE VALUE
      INTEGER    SEND_COUNTS(MAX_RANKS)
      INTEGER    SEND_DISPLS(MAX_RANKS)
      INTEGER    COUNTS(MAX_RANKS+MAX_RANKS)
      INTEGER    DISPLS(MAX_RANKS+MAX_RANKS)


      CHARACTER*(IOSIZE)  INFOBUF
      CHARACTER*32   TESTNAME
       CHARACTER*(MPI_MAX_ERROR_STRING) ERRORSTRING

 99   FORMAT(A,INT_FMT)
 179  FORMAT(A,I6,A,I6,A,I6,A,I6,A,I3,A,INT_FMT)
 189  FORMAT(INT_FMT,A,A,I6,A,I6,A,I6,A,I6,A,I3,A,INT_FMT)


!
!     Initialize the MPI environment
!
      CALL MPI_INIT(IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        WRITE(INFOBUF,99) 'MPI_INIT() returned', IERR
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF

!
!     Initialize the MPITEST environment
!
      CALL MPITEST_INIT(IERR)
      TESTNAME = 'MPI_Scatterv()'
      IF (MPITEST_ME .EQ. 0) THEN
        INFOBUF = 'Starting test '//TESTNAME
        CALL MPITEST_MESSAGE(MPITEST_INFO0, INFOBUF)
      END IF


!
!     Initialize the loop/error counters
!
      LOOP_CNT = 0
      FAIL = 0

!
!     Get number of communicators, data types and message lengths to test
!
      CALL MPITEST_NUM_COMM_SIZES(COMM_COUNT)
      CALL MPITEST_NUM_DATATYPES(TYPE_COUNT)
      CALL MPITEST_NUM_MESSAGE_LENGTHS(LENGTH_COUNT)

!
!     Get max message length for test
!
      CALL MPITEST_GET_MAX_MESSAGE_LENGTH(MAX_BYTE_LENGTH)

!
!     Loop through communicators
!
      DO 100 I=1,COMM_COUNT
        CALL MPITEST_GET_COMM_INDEX(I, COMM_INDEX)
        CALL MPITEST_GET_COMM_TYPE(I, COMM_TYPE)

        CALL MPITEST_GET_COMMUNICATOR(COMM_TYPE, COMM_INDEX, COMM, &
              COMM_SIZE)

!       Skip to end of loop if this node is not part of current
!       communicator

        IF (COMM .EQ. MPI_COMM_NULL) GO TO 201

!       Skip if intercommunicator

        CALL MPI_COMM_TEST_INTER(COMM, INTER_FLAG, IERR)
        IF (INTER_FLAG .EQV. .TRUE.) GO TO 201

!
!       Loop through data types.
!
        DO 200 J=1,TYPE_COUNT

          CALL MPITEST_GET_DATATYPE(J, TEST_TYPE)
!
!         Convert the number of bytes in the maximum length message
!         into the number of elements of the current type 
          CALL MPITEST_BYTE_TO_ELEMENT(TEST_TYPE, MAX_BYTE_LENGTH, &
                MAX_LENGTH)
          CALL MPITEST_BYTE_TO_ELEMENT(TEST_TYPE, &
                MAX_BUFF_SIZE * MPITEST_BUF_EXTENT, MAX_BUFF_LENGTH)

          IF (MAX_LENGTH .GE. (MAX_BUFF_LENGTH / MPITEST_NUMP) - 2) &
               MAX_LENGTH = (MAX_BUFF_LENGTH / MPITEST_NUMP) - 2
          IF (MAX_LENGTH .LE. 0) MAX_LENGTH = 0

!
!         Loop through data lengths
!
          DO 300 K=1,LENGTH_COUNT
            CALL MPITEST_GET_MESSAGE_LENGTH(K, BYTE_LENGTH)
            CALL MPITEST_BYTE_TO_ELEMENT(TEST_TYPE, BYTE_LENGTH, &
                LENGTH)
            IF (LENGTH .GT. MAX_LENGTH) LENGTH = MAX_LENGTH

            DO 600 ROOT=0, COMM_SIZE-1
!
!             Initialize send data
!
              LEADING_DIM=LENGTH + 2
              L = 0
              DO 700 M=1, COMM_SIZE
                DISPLS((2*M)-1) = (M-1) * LEADING_DIM
                DISPLS(2*M) = DISPLS(2*M-1)+LENGTH+L
                COUNTS((2*M)-1) = LENGTH + L
                COUNTS(2*M) = LEADING_DIM - LENGTH - L
                SEND_COUNTS(M) = LENGTH + L
                SEND_DISPLS(M) = (M-1) * LEADING_DIM
!                CALL MPITEST_DATATEMPLATE_INIT_POS(VALUES,
!     $              ROOT + M - 1, (2*M)-1, TEST_TYPE)
!                CALL MPITEST_DATATEMPLATE_INIT_POS(VALUES, -1, 2*M,
!     $               TEST_TYPE)

                CALL MPITEST_DATATEMPLATE_INIT_POS(VALUES, &
                    ROOT + M - 1, (2*M)-1, TEST_TYPE)
                CALL MPITEST_DATATEMPLATE_INIT_POS(VALUES, -1, 2*M, &
                     TEST_TYPE)
                L = 1 - L
 700          CONTINUE

              CALL MPITEST_INIT_BUFFER_V(TEST_TYPE, 2*COMM_SIZE, &
                    COUNTS, DISPLS, VALUES, SEND_BUFFER)
!
!             Initialize receive buffer
!
               CALL MPITEST_DATATEMPLATE_INIT(VALUE, -1, &
                      TEST_TYPE)
              CALL MPITEST_INIT_BUFFER(TEST_TYPE, LENGTH+2, &
                    VALUE, RECV_BUFFER)
              CALL MPITEST_DATATEMPLATE_INIT(VALUE, &
                    ROOT + MPITEST_CURRENT_RANK, TEST_TYPE)

!             Print an informational message
              IF (MPITEST_CURRENT_RANK .EQ. 0) THEN
                WRITE (INFOBUF,179) 'commindex',COMM_INDEX,', commsize', &
                    COMM_SIZE, ', commtype', COMM_TYPE, ', root', ROOT, &
                    ', datatype', TEST_TYPE, ', length',  LENGTH
                CALL MPITEST_MESSAGE(MPITEST_INFO1, INFOBUF)
              END IF

              LOOP_CNT = LOOP_CNT + 1

              L = MOD(MPITEST_CURRENT_RANK, 2)
              CALL MPI_SCATTERV(SEND_BUFFER, SEND_COUNTS, SEND_DISPLS, &
                   MPITEST_MPI_DATATYPES(TEST_TYPE), &
                   RECV_BUFFER, LENGTH + L,  &
                   MPITEST_MPI_DATATYPES(TEST_TYPE), ROOT, COMM, IERR)
              IF (IERR .NE. MPI_SUCCESS) THEN
                WRITE(INFOBUF,99) 'MPI_SCATTERV() returned', IERR
                CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
                INFOBUF=' '
                CALL MPI_ERROR_STRING(IERR, ERRORSTRING, ERRSIZE, ERR)
                CALL MPITEST_MESSAGE(MPITEST_NONFATAL, ERRORSTRING)
                FAIL = FAIL + 1
              END IF
!
!             Check received data for errors . . .
!
              CALL MPITEST_BUFFER_ERRORS(TEST_TYPE, LENGTH + L, &
                    VALUE, RECV_BUFFER, ERROR)
!             . . . and overflow
              CALL MPITEST_DATATEMPLATE_INIT(VALUE, -1, &
                      TEST_TYPE)
              CALL MPITEST_BUFFER_ERRORS_OV(TEST_TYPE, LENGTH + L, &
                    VALUE, RECV_BUFFER, ERROR2)
              
              IF (ERROR+ERROR2 .NE. 0) THEN
                IF (IERR .EQ. MPI_SUCCESS) FAIL = FAIL + 1
                WRITE (INFOBUF,189) ERROR+ERROR2, ' errors in buffer, ', &
                    'commindex',COMM_INDEX,', commsize', &
                    COMM_SIZE, ', commtype', COMM_TYPE, ', root', ROOT, &
                    ', datatype', TEST_TYPE, ', length', LENGTH
                CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
              END IF
 600        CONTINUE
                
 300      CONTINUE
      
 200    CONTINUE
 201    CONTINUE

        CALL MPITEST_FREE_COMMUNICATOR(COMM_TYPE, COMM, IERR)

 100  CONTINUE

!
!     Report overall results
!
      CALL MPITEST_REPORT(LOOP_CNT - FAIL, FAIL, 0, TESTNAME)
      
      CALL MPI_FINALIZE(IERR)

      END  
  
  
